//+------------------------------------------------------------------+
//|                                                        Stars.mq4 |
//|                        Copyright 2021, MetaQuotes Software Corp. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2021, Nikolay Semko"
#property link      "https://www.mql5.com/ru/users/nikolay7ko"
#property link      "SemkoNV@bk.ru"
#property version   "1.00"
#property indicator_chart_window
#include <Canvas\iCanvas_CB.mqh> 

#property indicator_chart_window
#property indicator_plots   1
#property indicator_buffers 1
  
input int NS = 10000; // Total stars
input bool AutoRotation = true; 
input double RotationSpeed = 1.1;
input double MaxSizeRadius = 0.4;

struct Star
  {
   double             x;
   double             y;
   double             z;
   double             size;
  };
Star st[];
double X0,Y0;
double _a,_b,_x,_y,z2,_r;
double per=1;
double K;
//+------------------------------------------------------------------+
int OnInit()
  {
   ChartSetInteger(0,CHART_FOREGROUND,true);
   ArrayResize(st,NS);
   //ChartSetInteger(0,CHART_SHOW,false);
   //Canvas.SetBack(true);
   StarArray(st);
   _a =1.0;
   _b = 1.0;
   per=20*_Height;
   K=100.0*_Height;
   _r = _Height*1.5;
   X0=_Width/2.0;
   Y0=_Height/2.0;

   Draw();
   if (AutoRotation) EventSetMillisecondTimer(20);
   return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
void OnTimer() {
   static int xm = _MouseX;
   static int ym = _MouseY;
   if ( W.MouseX != xm || W.MouseX != xm) {
      _a=2.0*(1.1*_Width  - _MouseX)/_Width;
      _b=1.0*(1.1*_Height - _MouseY)/_Height;
      xm = _MouseX;
      ym = _MouseY;
   } else {
      _a += 0.0001*RotationSpeed;
      _b += 0.00007*RotationSpeed; 
   }
   Draw();
}
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   //if(reason<2)
  //    ChartSetInteger(0,CHART_SHOW,true);
  }
//+------------------------------------------------------------------+
int OnCalculate(const int rates_total,const int prev_calculated,const int begin,const double &price[])
  {
   //Draw();
   return(rates_total);
  }
//+------------------------------------------------------------------+
void OnChartEvent(const int id,
                  const long &lparam,
                  const double &dparam,
                  const string &sparam)
  {
   if(id == CHARTEVENT_CHART_CHANGE) Draw();
  }
//+------------------------------------------------------------------+
void Draw()
  {
   //ulong t = GetMicrosecondCount();
   Canvas.Erase(0xFF000000);
   for(int i = 0; i<NS; i++)
     {
      _3D(st[i].x*_Width,st[i].y*_Height,st[i].z*_Height);
      iCirFill(_x,_y,st[i].size*(1.0+0.2*(z2/(1.5*_Height))),0xFFFFFFFF);
     }
   Canvas.Update();
   //t = GetMicrosecondCount() - t;
   //Comment(t);
  }
//+------------------------------------------------------------------+
void _3D(double x, double y, double z)
  {
   if(_a <0 || _b<0 || _a>=per || _b>=per)
      return;
   double x1=x*cos(_a)+z*sin(_a);
   double z1=-x*sin(_a)+z*cos(_a);
   double y1=y*cos(_b)+z1*sin(_b);
   z2=-y*sin(_b)+z1*cos(_b);
   z2+=_r;
   //Print(string(_Height*1.5) + "    " + string(z2));
   _x=(X0+K*x1/(z2+K));
   _y=(Y0+K*y1/(z2+K));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void StarArray(Star &arr[])
  {
   for(int i = 0; i<NS; i++)
     {
      arr[i].x = 3*rand()/32767.0 - 1.5;
      arr[i].y = 3*rand()/32767.0 - 1.5;
      arr[i].z = 3*rand()/32767.0 - 1.5;
      arr[i].size = MaxSizeRadius*rand()/32767.0;
     }
  }
//+------------------------------------------------------------------+
void iCirFill(double x,double y,double r, uint clr, double thick=1, double k=0.5)
  {
   int w=W.Width;
   double k2 = (r<0.5)?(r*r/0.25):1.0;
   for(int Xc=int(x-r); Xc<=int(x+r)+1; Xc++)
      for(int Yc=int(y-r); Yc<=int(y+r)+1; Yc++)
        {
         double d=sqrt((Xc-x)*(Xc-x)+(Yc-y)*(Yc-y));
         double D=d-r;
         if(D<thick+k)
           {
            double m=(D<thick-k)?k2:k2*(1-(D-thick+k)/(k*2));
            if(Yc>=0 && Yc<_Height && Xc>=0 && Xc<w)
               MixColor(clr,Canvas.m_pixels[Yc*w+Xc],m);
           }
        }
  }
  
void MixColor(uint clr, uint &addr, double mix) { 
   MixColor((uchar(Round(GETRGBA(clr)*mix+0.49999))<<24)|(clr&0x00FFFFFF),addr);
}

void MixColor(uint clr, uint &addr) { 
   uint clrback = addr;
   if (clrback == 0) {
      addr = clr;
      return;
   }
   argb C;
   C.clr = clr;
   if (C.c[3] == 0)
      return;
   if (C.c[3] == 255)
      addr = clr;
   argb Bg;
   Bg.clr = clrback;
   double mix = C.c[3] / 255.0;
   double mixb = Bg.c[3] / 255.0;
   C.c[2] = uchar(Bg.c[2] + mix * (C.c[2] - Bg.c[2]));
   C.c[1] = uchar(Bg.c[1] + mix * (C.c[1] - Bg.c[1]));
   C.c[0] = uchar(Bg.c[0] + mix * (C.c[0] - Bg.c[0]));
   C.c[3] = uchar((mixb + mix - mixb * mix) * 255 + 0.49999);
   addr = C.clr;
}

